"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.StandardRelayerApp = void 0;
const application_js_1 = require("./application.js");
const index_js_1 = require("./middleware/index.js");
const redis_storage_js_1 = require("./storage/redis-storage.js");
const utils_js_1 = require("./utils.js");
const logging_js_1 = require("./logging.js");
const bullMQAdapter_js_1 = require("@bull-board/api/bullMQAdapter.js");
const koa_1 = require("@bull-board/koa");
const api_1 = require("@bull-board/api");
const prom_client_1 = require("prom-client");
const defaultStdOpts = {
    spyEndpoint: "localhost:7073",
    workflows: {
        retries: 3,
    },
    fetchSourceTxhash: true,
    logger: logging_js_1.defaultLogger,
};
class StandardRelayerApp extends application_js_1.RelayerApp {
    store;
    mergedRegistry;
    constructor(env, opts) {
        // take logger out before merging because of recursive call stack
        const logger = opts.logger ?? logging_js_1.defaultLogger;
        delete opts.logger;
        // now we can merge
        const options = (0, utils_js_1.mergeDeep)({}, [
            defaultStdOpts,
            opts,
        ]);
        const { privateKeys, tokensByChain, name, spyEndpoint, redis, redisCluster, redisClusterEndpoints, wormholeRpcs, retryBackoffOptions, maxCompletedQueueSize, maxFailedQueueSize, } = options;
        super(env, options);
        this.store = new redis_storage_js_1.RedisStorage({
            redis,
            redisClusterEndpoints,
            redisCluster,
            attempts: options.workflows?.retries ?? 3,
            namespace: name,
            queueName: `${name}-relays`,
            exponentialBackoff: retryBackoffOptions,
            maxCompletedQueueSize,
            maxFailedQueueSize,
        });
        // this is always true for standard relayer app, but I'm adding this if
        // to make the dependency explicit
        if (this.store) {
            (0, index_js_1.spawnMissedVaaWorker)(this, {
                namespace: name,
                registry: this.store.registry,
                logger,
                redis,
                redisCluster,
                redisClusterEndpoints,
                wormholeRpcs,
                concurrency: options.missedVaaOptions?.concurrency,
                checkInterval: options.missedVaaOptions?.checkInterval,
                fetchVaaRetries: options.missedVaaOptions?.fetchVaaRetries,
                vaasFetchConcurrency: options.missedVaaOptions?.vaasFetchConcurrency,
                storagePrefix: this.store.getPrefix(),
                startingSequenceConfig: options.missedVaaOptions?.startingSequenceConfig,
                forceSeenKeysReindex: options.missedVaaOptions?.forceSeenKeysReindex,
            });
        }
        this.mergedRegistry = prom_client_1.Registry.merge([
            this.store.registry,
            super.metricsRegistry,
        ]);
        this.spy(spyEndpoint);
        this.useStorage(this.store);
        this.logger(logger);
        this.use((0, index_js_1.logging)(logger));
        this.use((0, index_js_1.providers)(options.providers, Object.keys(privateKeys ?? {})));
        // You need valid private keys to turn on the wallet middleware
        if (privateKeys !== undefined && Object.keys(privateKeys).length > 0) {
            this.use((0, index_js_1.wallets)(env, {
                logger,
                namespace: name,
                privateKeys,
                tokensByChain,
                metrics: { enabled: true, registry: this.metricsRegistry },
            }));
        }
        this.use((0, index_js_1.tokenBridgeContracts)());
        this.use((0, index_js_1.stagingArea)({
            namespace: name,
            redisCluster,
            redis,
            redisClusterEndpoints,
        }));
        if (options.fetchSourceTxhash) {
            this.use((0, index_js_1.sourceTx)());
        }
    }
    /**
     * Registry with prometheus metrics exported by the relayer.
     * Metrics include:
     * - active_workflows: Number of workflows currently running
     * - delayed_workflows: Number of worklows which are scheduled in the future either because they were scheduled that way or because they failed.
     * - waiting_workflows: Workflows waiting for a worker to pick them up.
     * - worklow_processing_duration: Processing time for completed jobs (processing until completed)
     * - workflow_total_duration: Processing time for completed jobs (processing until completed)
     */
    get metricsRegistry() {
        return this.mergedRegistry;
    }
    /**
     * A UI that you can mount in a KOA app to show the status of the queue / jobs.
     * @param path
     */
    storageKoaUI(path) {
        // UI
        const serverAdapter = new koa_1.KoaAdapter();
        serverAdapter.setBasePath(path);
        (0, api_1.createBullBoard)({
            queues: [new bullMQAdapter_js_1.BullMQAdapter(this.store.vaaQueue)],
            serverAdapter: serverAdapter,
        });
        return serverAdapter.registerPlugin();
    }
}
exports.StandardRelayerApp = StandardRelayerApp;
//# sourceMappingURL=application-standard.js.map