"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.tryFetchVaa = exports.updateMetrics = exports.calculateSequenceStats = void 0;
const fail_fast_grpc_transport_js_1 = require("../../rpc/fail-fast-grpc-transport.js");
const wormhole_sdk_1 = require("@certusone/wormhole-sdk");
function calculateSequenceStats(runStats, failedToFetchSequences, previousSafeSequence) {
    const { seenSequences } = runStats;
    const lastSafeSequence = calculateLastSafeSequence(runStats, failedToFetchSequences, previousSafeSequence);
    const lastSeenSequence = Math.max(seenSequences.length ? Number(seenSequences[seenSequences.length - 1]) : 0, lastSafeSequence);
    const firstSeenSequence = Number(seenSequences[0]) || 0;
    return { lastSafeSequence, lastSeenSequence, firstSeenSequence };
}
exports.calculateSequenceStats = calculateSequenceStats;
function calculateLastSafeSequence(runStats, failedToFetchSequences, previousSafeSequence) {
    if (failedToFetchSequences && failedToFetchSequences.length > 0) {
        // we have sequences that we have failed to update before. We won't update the last
        // safe sequence. Return the the last sequence before the first one we failed to fetch.
        return Number(failedToFetchSequences[0]) - 1;
    }
    const { missingSequences, failedToRecover, failedToReprocess } = runStats;
    const missingSequencesFailedToReprocess = failedToRecover.length > 0 || failedToReprocess.length > 0;
    if (missingSequences.length > 0 && missingSequencesFailedToReprocess) {
        // we found some missing sequences on this run, but we were able to reprocess them
        // return the previous safe sequence, or 0 if there is none.
        return Number(missingSequences[0]) - 1;
    }
    // No missing sequences up to seenSequences
    // if the there was vaas recovered by the lookahead, use that as the
    // last safe sequence. Otherwise, use the last seen sequence.
    const lastSeenSequence = runStats.lookAheadSequences.length > 0
        ? runStats.lookAheadSequences[runStats.lookAheadSequences.length - 1]
        : runStats.seenSequences[runStats.seenSequences.length - 1];
    return lastSeenSequence
        ? Number(lastSeenSequence)
        : previousSafeSequence
            ? Number(previousSafeSequence)
            : 0;
}
function updateMetrics(metrics, filter, startTime, // timestamp in ms
failure, missedVaas, sequenceStats) {
    const { emitterChain, emitterAddress } = filter;
    if (failure) {
        metrics.workerFailedRuns?.labels().inc();
        metrics.workerRunDuration?.labels().observe(Date.now() - startTime);
        return;
    }
    metrics.workerSuccessfulRuns?.labels().inc();
    metrics.workerRunDuration?.labels().observe(Date.now() - startTime);
    const { processed, failedToRecover, lookAheadSequences, missingSequences, failedToReprocess, } = missedVaas;
    const vaasProcessed = processed.length;
    // This are VAAs that were found missing between known sequences, but we failed
    // to fetch them to reprocess them
    const vaasFailedToRecover = failedToRecover.length;
    // This are VAAs that were found but failed when trying to re-queue them
    const vaasFailedToReprocess = failedToReprocess.length;
    const vaasFound = missingSequences.length + lookAheadSequences.length;
    const labels = {
        emitterChain: (0, wormhole_sdk_1.coalesceChainName)(emitterChain),
        emitterAddress,
    };
    if (vaasFound > 0) {
        metrics.detectedVaas?.labels(labels).inc(vaasFound);
    }
    if (vaasProcessed > 0) {
        metrics.recoveredVaas?.labels(labels).inc(vaasProcessed);
    }
    if (vaasFailedToRecover > 0) {
        metrics.failedToRecover?.labels(labels).inc(vaasFailedToRecover);
    }
    if (vaasFailedToReprocess > 0) {
        metrics.failedToReprocess?.labels(labels).inc(vaasFailedToReprocess);
    }
    const { lastSeenSequence, lastSafeSequence, firstSeenSequence } = sequenceStats;
    metrics.lastSeenSequence?.labels(labels).set(lastSeenSequence);
    metrics.lastSafeSequence?.labels(labels).set(lastSafeSequence);
    metrics.firstSeenSequence?.labels(labels).set(firstSeenSequence);
    metrics.missingSequences?.labels(labels).set(missingSequences.length ? 1 : 0);
}
exports.updateMetrics = updateMetrics;
async function tryFetchVaa(vaaKey, wormholeRpcs, retries = 2) {
    let vaa;
    const stack = new Error().stack;
    try {
        vaa = await (0, wormhole_sdk_1.getSignedVAAWithRetry)(wormholeRpcs, vaaKey.emitterChain, vaaKey.emitterAddress, vaaKey.sequence, { transport: (0, fail_fast_grpc_transport_js_1.FailFastGrpcTransportFactory)() }, 100, retries * wormholeRpcs.length);
    }
    catch (error) {
        error.stack = new Error().stack;
        if (error.code === 5) {
            return null;
        }
        throw error;
    }
    return vaa;
}
exports.tryFetchVaa = tryFetchVaa;
//# sourceMappingURL=helpers.js.map