"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getAllProcessedSeqsInOrder = exports.calculateStartingIndex = exports.tryGetExistingFailedSequences = exports.tryGetLastSafeSequence = exports.trySetLastSafeSequence = exports.updateSeenSequences = exports.deleteExistingSeenVAAsData = exports.batchMarkAsFailedToRecover = exports.batchMarkAsSeen = exports.markVaaAsSeen = void 0;
async function markVaaAsSeen(redis, vaaKey, prefix) {
    const { emitterChain, emitterAddress, sequence } = vaaKey;
    const seenVaaKey = getSeenVaaKey(prefix, emitterChain, emitterAddress);
    const sequencesString = sequence.toString();
    await redis.zadd(seenVaaKey, sequencesString, sequencesString);
}
exports.markVaaAsSeen = markVaaAsSeen;
function batchMarkAsSeen(redis, prefix, emitterChain, emitterAddress, sequences) {
    return Promise.all([
        batchAddToSet(redis, getSeenVaaKey(prefix, emitterChain, emitterAddress), sequences),
        batchRemoveFromSet(redis, getFailedToFetchKey(prefix, emitterChain, emitterAddress), sequences),
    ]);
}
exports.batchMarkAsSeen = batchMarkAsSeen;
function batchMarkAsFailedToRecover(redis, prefix, emitterChain, emitterAddress, sequences) {
    return batchAddToSet(redis, getFailedToFetchKey(prefix, emitterChain, emitterAddress), sequences);
}
exports.batchMarkAsFailedToRecover = batchMarkAsFailedToRecover;
async function deleteExistingSeenVAAsData(filters, redis, storagePrefix, logger) {
    logger?.info("Deleting existing VAAs and failed VAAs. Will recreate index from redis-storage");
    const pipeline = redis.pipeline();
    for (const filter of filters) {
        pipeline.del(getSeenVaaKey(storagePrefix, filter.emitterChain, filter.emitterAddress));
        pipeline.del(getFailedToFetchKey(storagePrefix, filter.emitterChain, filter.emitterAddress));
        pipeline.del(getSafeSequenceKey(storagePrefix, filter.emitterChain, filter.emitterAddress));
    }
    await pipeline.exec();
}
exports.deleteExistingSeenVAAsData = deleteExistingSeenVAAsData;
async function updateSeenSequences(filters, redis, storagePrefix) {
    let scannedKeys = 0;
    for (const filter of filters) {
        const seenVaaKey = getSeenVaaKey(storagePrefix, filter.emitterChain, filter.emitterAddress);
        scannedKeys += await scanNextBatchAndUpdateSeenSequences(redis, filter, storagePrefix, seenVaaKey);
    }
    return scannedKeys;
}
exports.updateSeenSequences = updateSeenSequences;
async function trySetLastSafeSequence(redis, prefix, emitterChain, emitterAddress, lastSeenSequence, logger) {
    const key = getSafeSequenceKey(prefix, emitterChain, emitterAddress);
    // since safe sequence is not critical, we'll swallow the error
    try {
        await redis.set(key, lastSeenSequence);
    }
    catch (error) {
        logger?.warn(`Error setting last safe sequence for chain: ${emitterChain}`, error);
        return false;
    }
    return true;
}
exports.trySetLastSafeSequence = trySetLastSafeSequence;
async function tryGetLastSafeSequence(redis, prefix, emitterChain, emitterAddress, logger) {
    // since safe sequence is not critical, we'll swallow the error
    const key = getSafeSequenceKey(prefix, emitterChain, emitterAddress);
    let lastSafeSequence;
    try {
        lastSafeSequence = await redis.get(key);
    }
    catch (error) {
        logger?.warn(`Error getting last safe sequence for chain: ${emitterChain}`, error);
        return undefined;
    }
    return lastSafeSequence !== null && lastSafeSequence !== undefined
        ? BigInt(lastSafeSequence)
        : undefined;
}
exports.tryGetLastSafeSequence = tryGetLastSafeSequence;
async function tryGetExistingFailedSequences(redis, filter, prefix) {
    const failedToFetchKey = getFailedToFetchKey(prefix, filter.emitterChain, filter.emitterAddress);
    let failedToFetchSequences;
    try {
        failedToFetchSequences = await getDataFromSortedSet(redis, failedToFetchKey);
    }
    catch (error) {
        return error;
    }
    return failedToFetchSequences;
}
exports.tryGetExistingFailedSequences = tryGetExistingFailedSequences;
async function calculateStartingIndex(redis, prefix, emitterChain, emitterAddress, lastSafeSequence) {
    const key = getSeenVaaKey(prefix, emitterChain, emitterAddress);
    let indexToStartFrom = null;
    if (lastSafeSequence !== undefined) {
        indexToStartFrom = await redis.zrank(key, lastSafeSequence.toString());
    }
    return indexToStartFrom;
}
exports.calculateStartingIndex = calculateStartingIndex;
async function getAllProcessedSeqsInOrder(redis, prefix, emitterChain, emitterAddress, indexToStartFrom, startingSequenceConfig) {
    const key = getSeenVaaKey(prefix, emitterChain, emitterAddress);
    const results = await getDataFromSortedSet(redis, key, indexToStartFrom?.toString());
    const orderedResults = results
        .map(r => Number(r))
        .sort((a, b) => (a > b ? 1 : -1))
        .map(BigInt);
    if (startingSequenceConfig && startingSequenceConfig > orderedResults[0]) {
        const higherSequences = orderedResults.filter(r => r >= startingSequenceConfig);
        if (higherSequences[0] > startingSequenceConfig) {
            higherSequences.unshift(BigInt(startingSequenceConfig) - 1n);
        }
        return higherSequences;
    }
    return orderedResults;
}
exports.getAllProcessedSeqsInOrder = getAllProcessedSeqsInOrder;
/**
 *
 * Private Functions:
 *
 */
function getSeenVaaKey(prefix, emitterChain, emitterAddress) {
    return `${prefix}:missedVaasV3:seenVaas:${emitterChain}:${emitterAddress}`;
}
function getFailedToFetchKey(prefix, emitterChain, emitterAddress) {
    return `${prefix}:missedVaasV3:failedToFetch:${emitterChain}:${emitterAddress}`;
}
function getSafeSequenceKey(prefix, emitterChain, emitterAddress) {
    return `${prefix}:missedVaasV3:safeSequence:${emitterChain}:${emitterAddress}`;
}
// example keys:
// {GenericRelayer}:GenericRelayer-relays:14/000000000000000000000000306b68267deb7c5dfcda3619e22e9ca39c374f84/55/O8xvv:logs
// {GenericRelayer}:GenericRelayer-relays:14/000000000000000000000000306b68267deb7c5dfcda3619e22e9ca39c374f84/49/d1Cjd
function parseStorageVaaKey(key) {
    const vaaIdString = key.split(":")[2];
    const sequenceString = vaaIdString.split("/")[2];
    return sequenceString;
}
const CURSOR_IDENTIFIER = "0";
async function scanNextBatchAndUpdateSeenSequences(redis, filter, storagePrefix, seenVaaKey, cursor = CURSOR_IDENTIFIER) {
    const { emitterChain, emitterAddress } = filter;
    const prefix = `${storagePrefix}:${emitterChain}/${emitterAddress}`;
    const [nextCursor, keysFound] = await redis.scan(cursor, "MATCH", `${prefix}*`);
    const pipeline = redis.pipeline();
    let scannedKeys = 0;
    for (const key of keysFound) {
        scannedKeys++;
        if (key.endsWith(":logs"))
            continue;
        const vaaSequence = parseStorageVaaKey(key);
        pipeline.zadd(seenVaaKey, vaaSequence, vaaSequence);
    }
    await pipeline.exec();
    let nextBatchScannedKeys = 0;
    if (nextCursor !== CURSOR_IDENTIFIER) {
        nextBatchScannedKeys = await scanNextBatchAndUpdateSeenSequences(redis, filter, storagePrefix, seenVaaKey, nextCursor);
    }
    return scannedKeys + nextBatchScannedKeys;
}
function getDataFromSortedSet(redis, key, lowerBound) {
    const lb = lowerBound || "0";
    return redis.zrange(key, lb, -1);
}
async function batchAddToSet(redis, key, sequences) {
    const pipeline = redis.pipeline();
    for (const sequence of sequences) {
        pipeline.zadd(key, sequence, sequence);
    }
    await pipeline.exec();
}
async function batchRemoveFromSet(redis, key, sequences) {
    const pipeline = redis.pipeline();
    for (const sequence of sequences) {
        pipeline.zrem(key, sequence);
    }
    return pipeline.exec();
}
//# sourceMappingURL=storage.js.map