"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.providers = void 0;
const wormhole_sdk_1 = require("@certusone/wormhole-sdk");
const ethers_1 = require("ethers");
const solana = require("@solana/web3.js");
const consts_js_1 = require("@certusone/wormhole-sdk/lib/cjs/utils/consts.js");
const sui = require("@mysten/sui.js");
const environment_js_1 = require("../environment.js");
const core_1 = require("@sei-js/core");
const utils_js_1 = require("../utils.js");
const defaultSupportedChains = {
    [environment_js_1.Environment.MAINNET]: {
        [wormhole_sdk_1.CHAIN_ID_SOLANA]: { endpoints: ["https://api.mainnet-beta.solana.com"] },
        [wormhole_sdk_1.CHAIN_ID_ETH]: { endpoints: ["https://rpc.ankr.com/eth"] },
        [wormhole_sdk_1.CHAIN_ID_BSC]: { endpoints: ["https://bsc-dataseed1.binance.org/"] },
        [consts_js_1.CHAIN_ID_POLYGON]: { endpoints: ["https://rpc.ankr.com/polygon"] },
        [consts_js_1.CHAIN_ID_AVAX]: { endpoints: ["https://api.avax.network/ext/bc/C/rpc"] },
        [consts_js_1.CHAIN_ID_FANTOM]: { endpoints: ["https://rpc.ftm.tools"] },
        [wormhole_sdk_1.CHAIN_ID_CELO]: { endpoints: ["https://forno.celo.org"] },
        [wormhole_sdk_1.CHAIN_ID_MOONBEAM]: { endpoints: ["https://rpc.api.moonbeam.network"] },
        [wormhole_sdk_1.CHAIN_ID_ACALA]: { endpoints: ["https://eth-rpc-acala.aca-api.network"] },
        [wormhole_sdk_1.CHAIN_ID_ALGORAND]: { endpoints: ["https://node.algoexplorerapi.io/"] },
        [consts_js_1.CHAIN_ID_ARBITRUM]: {
            endpoints: [
                "https://arbitrum-one.publicnode.com",
                "https://rpc.ankr.com/arbitrum",
            ],
        },
        [wormhole_sdk_1.CHAIN_ID_APTOS]: {
            endpoints: ["https://fullnode.mainnet.aptoslabs.com/v1"],
        },
        [wormhole_sdk_1.CHAIN_ID_SUI]: {
            endpoints: ["https://rpc.mainnet.sui.io:443"],
            faucets: [""],
            websockets: [""],
        },
        [consts_js_1.CHAIN_ID_KLAYTN]: {
            endpoints: ["https://public-node-api.klaytnapi.com/v1/cypress"],
        },
        [consts_js_1.CHAIN_ID_OPTIMISM]: {
            endpoints: ["https://optimism.api.onfinality.io/public"],
        },
        [wormhole_sdk_1.CHAIN_ID_BASE]: {
            endpoints: ["https://developer-access-mainnet.base.org"],
        },
    },
    [environment_js_1.Environment.TESTNET]: {
        // [CHAIN_ID_ALGORAND]: { endpoints: ["node.testnet.algoexplorerapi.io/"] },
        [wormhole_sdk_1.CHAIN_ID_SOLANA]: {
            endpoints: ["https://api.devnet.solana.com"],
        },
        [wormhole_sdk_1.CHAIN_ID_ETH]: {
            endpoints: [
                "https://eth-goerli.g.alchemy.com/v2/mvFFcUhFfHujAOewWU8kH5D1R2bgFgLt",
            ],
        },
        [wormhole_sdk_1.CHAIN_ID_BSC]: {
            endpoints: ["https://data-seed-prebsc-1-s3.binance.org:8545"],
        },
        [consts_js_1.CHAIN_ID_POLYGON]: {
            endpoints: ["https://matic-mumbai.chainstacklabs.com"],
        },
        [consts_js_1.CHAIN_ID_AVAX]: {
            endpoints: ["https://api.avax-test.network/ext/bc/C/rpc"],
        },
        [consts_js_1.CHAIN_ID_FANTOM]: { endpoints: ["https://rpc.ankr.com/fantom_testnet"] },
        [wormhole_sdk_1.CHAIN_ID_CELO]: {
            endpoints: ["https://alfajores-forno.celo-testnet.org"],
        },
        [wormhole_sdk_1.CHAIN_ID_MOONBEAM]: {
            endpoints: ["https://rpc.testnet.moonbeam.network"],
        },
        [wormhole_sdk_1.CHAIN_ID_APTOS]: {
            endpoints: ["https://fullnode.devnet.aptoslabs.com/v1"],
        },
        [wormhole_sdk_1.CHAIN_ID_SUI]: {
            endpoints: [sui.testnetConnection.fullnode],
            faucets: [sui.testnetConnection.faucet],
            websockets: [sui.testnetConnection.websocket],
        },
        [wormhole_sdk_1.CHAIN_ID_SEI]: {
            endpoints: ["https://sei-testnet-2-rpc.brocha.in"],
        },
        [consts_js_1.CHAIN_ID_KLAYTN]: {
            endpoints: ["https://public-en-cypress.klaytn.net"],
        },
        [consts_js_1.CHAIN_ID_OPTIMISM]: {
            endpoints: ["https://goerli.optimism.io"],
        },
        [consts_js_1.CHAIN_ID_ARBITRUM]: {
            endpoints: [
                "https://arbitrum-goerli.public.blastapi.io",
                "https://arbitrum-goerli.publicnode.com",
            ],
        },
        [wormhole_sdk_1.CHAIN_ID_BASE]: {
            endpoints: ["https://goerli.base.org"],
        },
    },
    [environment_js_1.Environment.DEVNET]: {
        [wormhole_sdk_1.CHAIN_ID_ETH]: {
            endpoints: ["http://localhost:8545/"],
        },
        [wormhole_sdk_1.CHAIN_ID_BSC]: {
            endpoints: ["http://localhost:8546/"],
        },
    },
};
function pick(obj, keys) {
    const res = {};
    for (const key of keys) {
        if (key in obj) {
            // We need to assert that this is a key of T because `key in obj` does not provide this type guard
            const keyAsserted = key;
            res[keyAsserted] = obj[keyAsserted];
        }
    }
    return res;
}
/**
 * providers is a middleware that populates `ctx.providers` with provider information
 * @param opts
 */
function providers(opts, supportedChains) {
    let providers;
    return async (ctx, next) => {
        const logger = ctx.logger?.child({ module: "providers" });
        if (!providers) {
            // it makes no sense to start providers for all default
            // chains, we should only start providers for the chains the relayer
            // will use.
            // The config is optional because of backwards compatibility
            // If no config is passed, we'll start providers for all default chains
            const environmentDefaultSupportedChains = defaultSupportedChains[ctx.env];
            const defaultChains = supportedChains
                ? pick(environmentDefaultSupportedChains, supportedChains)
                : environmentDefaultSupportedChains;
            const chains = Object.assign({}, defaultChains, opts?.chains);
            logger?.debug(`Providers initializing... ${JSON.stringify(maskRPCProviders(chains))}`);
            providers = await buildProviders(chains, logger);
            logger?.debug(`Providers initialized succesfully.`);
        }
        ctx.providers = providers;
        await next();
    };
}
exports.providers = providers;
const evmChainIds = wormhole_sdk_1.EVMChainNames.map(n => wormhole_sdk_1.CHAINS[n]);
const isEvmChainId = Object.fromEntries(evmChainIds.map(id => [id, true]));
const isEVMChain = (chainId) => isEvmChainId[chainId] ?? false;
async function buildProviders(supportedChains, logger) {
    const providers = {
        evm: {},
        solana: [],
        sui: [],
        sei: [],
        untyped: {},
    };
    for (const [chainIdStr, chainCfg] of Object.entries(supportedChains)) {
        const chainId = Number(chainIdStr);
        const { endpoints, faucets, websockets } = chainCfg;
        try {
            if (isEVMChain(chainId)) {
                providers.evm[chainId] = endpoints.map(url => new ethers_1.ethers.providers.JsonRpcProvider(url));
            }
            else if (chainId === wormhole_sdk_1.CHAIN_ID_SOLANA) {
                providers.solana = endpoints.map(url => new solana.Connection(url));
            }
            else if (chainId === wormhole_sdk_1.CHAIN_ID_SUI) {
                providers.sui = endpoints.map((url, i) => {
                    let conn = new sui.Connection({
                        fullnode: url,
                        faucet: faucets?.at(i) || faucets?.at(0), // try to map to the same index, otherwise use the first (if user only provided one faucet but multiple endpoints)
                        websocket: websockets?.at(i) || websockets?.at(0), // same as above
                    });
                    return new sui.JsonRpcProvider(conn);
                });
            }
            else if (chainId === wormhole_sdk_1.CHAIN_ID_SEI) {
                const seiProviderPromises = endpoints.map(url => (0, core_1.getCosmWasmClient)(url));
                providers.sei = await Promise.all(seiProviderPromises);
            }
            else {
                providers.untyped[chainId] = endpoints.map(c => ({ rpcUrl: c }));
            }
        }
        catch (error) {
            if (error instanceof Error) {
                error.originalStack = error.stack;
                Error.captureStackTrace(error);
            }
            logger?.error(`Failed to initialize provider for chain: ${chainIdStr} - endpoints: ${maskRPCEndpoints(endpoints)}. Error: ${(0, utils_js_1.printError)(error)}`);
            throw error;
        }
    }
    return providers;
}
function maskRPCEndpoints(endpoints) {
    return endpoints.map((url) => {
        const apiKeyPos = url.indexOf("apiKey");
        if (apiKeyPos > -1) {
            // Found API key in the RPC url, show only initial 3 chars and mask the rest
            return url.substring(0, apiKeyPos + 10) + "***";
        }
        return url;
    });
}
function maskRPCProviders(chains) {
    const maskedChains = {};
    for (const [chainId, chainConfig] of Object.entries(chains)) {
        maskedChains[chainId] = {
            endpoints: maskRPCEndpoints(chainConfig.endpoints),
        };
    }
    return maskedChains;
}
//# sourceMappingURL=providers.middleware.js.map