"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.fetchVaaHash = exports.sourceTx = void 0;
const wormhole_sdk_1 = require("@certusone/wormhole-sdk");
const environment_js_1 = require("../environment.js");
const lru_cache_1 = require("lru-cache");
const application_js_1 = require("../application.js");
const wormholescan_client_js_1 = require("../rpc/wormholescan-client.js");
const utils_js_1 = require("../utils.js");
const defaultOptsByEnv = {
    [environment_js_1.Environment.MAINNET]: {
        wormscanEndpoint: application_js_1.defaultWormscanUrl[environment_js_1.Environment.MAINNET],
        retries: 5,
        initialDelay: 1000,
        maxDelay: 45000,
        timeout: 5000,
    },
    [environment_js_1.Environment.TESTNET]: {
        wormscanEndpoint: application_js_1.defaultWormscanUrl[environment_js_1.Environment.TESTNET],
        retries: 3,
        initialDelay: 1000,
        maxDelay: 30000,
        timeout: 3000,
    },
    [environment_js_1.Environment.DEVNET]: {
        wormscanEndpoint: application_js_1.defaultWormscanUrl[environment_js_1.Environment.DEVNET],
        retries: 3,
        initialDelay: 500,
        maxDelay: 10000,
        timeout: 2000,
    },
};
function ifVAAFinalized(vaa) {
    const { consistencyLevel, emitterChain } = vaa;
    if (emitterChain === wormhole_sdk_1.CHAIN_ID_SOLANA) {
        return consistencyLevel === 32;
    }
    else if (emitterChain === wormhole_sdk_1.CHAIN_ID_BSC) {
        return consistencyLevel > 15;
    }
    return consistencyLevel !== 200 && consistencyLevel !== 201;
}
let wormholescan;
function sourceTx(optsWithoutDefaults) {
    let opts;
    const alreadyFetchedHashes = new lru_cache_1.LRUCache({ max: 1000 });
    return async (ctx, next) => {
        if (!opts) {
            // initialize options now that we know the environment from context
            opts = {
                ...defaultOptsByEnv[ctx.env],
                ...optsWithoutDefaults,
            };
        }
        if (ctx.vaa === undefined) {
            ctx.logger?.debug("Didn't get a VAA id. Skipping tx hash fetch.");
            await next();
            return;
        }
        const vaaId = `${ctx.vaa.id.emitterChain}-${ctx.vaa.id.emitterAddress}-${ctx.vaa.id.sequence}`;
        const txHashFromCache = alreadyFetchedHashes.get(vaaId);
        if (txHashFromCache) {
            ctx.logger?.debug(`Already fetched tx hash: ${txHashFromCache}`);
            ctx.sourceTxHash = txHashFromCache;
            await next();
            return;
        }
        const { emitterChain, emitterAddress, sequence } = ctx.vaa;
        ctx.logger?.debug("Fetching tx hash...");
        let txHash = await fetchVaaHash(emitterChain, emitterAddress, sequence, ctx.env, ctx.logger, opts);
        if (txHash === "") {
            ctx.logger?.debug("Could not retrieve tx hash.");
        }
        else {
            // TODO look at consistency level before using cache? (not sure what the checks are)
            if (ifVAAFinalized(ctx.vaa)) {
                alreadyFetchedHashes.set(vaaId, txHash);
            }
            ctx.logger?.debug(`Retrieved tx hash: ${txHash}`);
        }
        ctx.sourceTxHash = txHash;
        await next();
    };
}
exports.sourceTx = sourceTx;
async function fetchVaaHash(emitterChain, emitterAddress, sequence, env, logger, sourceTxOpts) {
    if (!wormholescan) {
        const opts = sourceTxOpts || defaultOptsByEnv[env];
        wormholescan = new wormholescan_client_js_1.WormholescanClient(new URL(opts.wormscanEndpoint), {
            retries: opts.retries,
            initialDelay: opts.initialDelay,
            maxDelay: opts.maxDelay,
            timeout: opts.timeout,
        });
    }
    const response = await wormholescan.getVaa(emitterChain, emitterAddress.toString("hex"), sequence);
    if ("error" in response) {
        logger?.error(`Error fetching tx hash: ${(0, utils_js_1.printError)(response.error)}`);
        throw response.error;
    }
    let txHash = response.data.txHash || "";
    if ((0, wormhole_sdk_1.isEVMChain)(emitterChain) &&
        txHash &&
        !txHash.startsWith("0x")) {
        txHash = `0x${txHash}`;
    }
    logger?.debug("Source Transaction Hash: " + txHash || "Not Found");
    return txHash;
}
exports.fetchVaaHash = fetchVaaHash;
//# sourceMappingURL=source-tx.middleware.js.map