"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.wallets = void 0;
const wormhole_sdk_1 = require("@certusone/wormhole-sdk");
const wallet_worker_js_1 = require("./wallet.worker.js");
const queue_1 = require("@datastructures-js/queue");
const wallet_management_js_1 = require("./wallet-management.js");
function makeExecuteFunc(actionQueues, pluginName, logger) {
    // push action onto actionQueue and have worker reject or resolve promise
    const func = (chainId, f) => {
        return new Promise((resolve, reject) => {
            const maybeQueue = actionQueues.get(chainId);
            if (!maybeQueue) {
                logger?.error(`Error making execute function. Unsupported chain: ${chainId}`);
                return reject("Chain not supported");
            }
            maybeQueue.enqueue({
                action: { chainId, f },
                pluginName,
                resolve,
                reject,
            });
        });
    };
    func.onSolana = (f) => func(wormhole_sdk_1.CHAIN_ID_SOLANA, f);
    func.onSui = (f) => func(wormhole_sdk_1.CHAIN_ID_SUI, f);
    func.onSei = (f) => func(wormhole_sdk_1.CHAIN_ID_SEI, f);
    func.onEVM = (chainId, f) => func(chainId, f);
    return func;
}
function wallets(env, opts) {
    const workerInfoMap = new Map(Object.entries(opts.privateKeys).map(([chainIdStr, keys]) => {
        //TODO update for all ecosystems
        let chainId = Number(chainIdStr);
        const workerInfos = keys.map((key, id) => ({
            id,
            targetChainId: chainId,
            targetChainName: wormhole_sdk_1.CHAIN_ID_TO_NAME[chainId],
            walletPrivateKey: key,
        }));
        return [chainId, workerInfos];
    }));
    if (opts.metrics) {
        (0, wallet_management_js_1.startWalletManagement)(env, opts.privateKeys, opts.tokensByChain, opts.metrics, opts.logger);
    }
    let executeFunction;
    return async (ctx, next) => {
        if (!executeFunction) {
            ctx.logger?.debug(`Initializing wallets...`);
            const actionQueues = new Map();
            for (const [chain, workerInfos] of workerInfoMap.entries()) {
                const actionQueue = new queue_1.Queue();
                actionQueues.set(chain, actionQueue);
                workerInfos.forEach(info => (0, wallet_worker_js_1.spawnWalletWorker)(actionQueue, ctx.providers, info, opts.logger));
            }
            executeFunction = makeExecuteFunc(actionQueues, opts.namespace ?? "default", opts.logger);
            ctx.logger?.debug(`Initialized wallets`);
        }
        ctx.logger?.debug("wallets attached to context");
        ctx.wallets = executeFunction;
        await next();
    };
}
exports.wallets = wallets;
//# sourceMappingURL=wallet.middleware.js.map