"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HttpClientError = exports.HttpClient = void 0;
const promises_1 = require("timers/promises");
const utils_js_1 = require("../utils.js");
/**
 * A simple HTTP client with exponential backoff retries and 429 handling.
 */
class HttpClient {
    initialDelay = 1000;
    maxDelay = 60000;
    retries = 0;
    timeout = 5000;
    cache = "default";
    constructor(options) {
        options?.initialDelay && (this.initialDelay = options.initialDelay);
        options?.maxDelay && (this.maxDelay = options.maxDelay);
        options?.retries && (this.retries = options.retries);
        options?.timeout && (this.timeout = options.timeout);
        options?.cache && (this.cache = options.cache);
    }
    async get(url, opts) {
        return this.executeWithRetry(url, "GET", opts);
    }
    async execute(url, method, opts) {
        let response;
        try {
            response = await fetch(url, {
                method: method,
                signal: AbortSignal.timeout(opts?.timeout ?? this.timeout),
                cache: opts?.cache ?? this.cache,
            });
        }
        catch (err) {
            // Connection / timeout error:
            throw new HttpClientError((0, utils_js_1.printError)(err));
        }
        if (!response.ok) {
            throw new HttpClientError(undefined, response, await response.json());
        }
        return await response.json();
    }
    async executeWithRetry(url, method, opts) {
        const maxRetries = opts?.retries ?? this.retries;
        let retries = 0;
        const initialDelay = opts?.initialDelay ?? this.initialDelay;
        const maxDelay = opts?.maxDelay ?? this.maxDelay;
        while (maxRetries >= 0) {
            try {
                return await this.execute(url, method, opts);
            }
            catch (err) {
                if (err instanceof HttpClientError) {
                    if (retries < maxRetries) {
                        const retryAfter = err.getRetryAfter(maxDelay, err);
                        if (retryAfter) {
                            await (0, promises_1.setTimeout)(retryAfter, { ref: false });
                        }
                        else {
                            const timeout = Math.min(initialDelay * 2 ** maxRetries, maxDelay);
                            await (0, promises_1.setTimeout)(timeout, { ref: false });
                        }
                        retries++;
                        continue;
                    }
                }
                throw err;
            }
        }
        throw new Error("Failed to execute HTTP request.");
    }
}
exports.HttpClient = HttpClient;
class HttpClientError extends Error {
    status;
    data;
    headers;
    constructor(message, response, data) {
        super(message ?? `Unexpected status code: ${response?.status}`);
        this.status = response?.status;
        this.data = data;
        this.headers = response?.headers;
        Error.captureStackTrace(this, this.constructor);
    }
    /**
     * Parses the Retry-After header and returns the value in milliseconds.
     * @param maxDelay
     * @param error
     * @throws {HttpClientError} if retry-after is bigger than maxDelay.
     * @returns the retry-after value in milliseconds.
     */
    getRetryAfter(maxDelay, error) {
        const retryAfter = this.headers?.get("Retry-After");
        if (retryAfter) {
            const value = parseInt(retryAfter) * 1000; // header value is in seconds
            if (value <= maxDelay) {
                return value;
            }
            throw error;
        }
    }
}
exports.HttpClientError = HttpClientError;
//# sourceMappingURL=http-client.js.map