"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RedisStorage = void 0;
const bullmq_1 = require("bullmq");
const wormhole_sdk_1 = require("@certusone/wormhole-sdk");
const ioredis_1 = require("ioredis");
const storage_metrics_js_1 = require("../storage.metrics.js");
const utils_js_1 = require("../utils.js");
const koa_1 = require("@bull-board/koa");
const api_1 = require("@bull-board/api");
const bullMQAdapter_js_1 = require("@bull-board/api/bullMQAdapter.js");
function serializeVaa(vaa) {
    return {
        sequence: vaa.sequence.toString(),
        hash: vaa.hash.toString("base64"),
        emitterChain: vaa.emitterChain,
        emitterAddress: vaa.emitterAddress.toString("hex"),
        payload: vaa.payload.toString("base64"),
        nonce: vaa.nonce,
        timestamp: vaa.timestamp,
        version: vaa.version,
        guardianSignatures: vaa.guardianSignatures.map(sig => ({
            signature: sig.signature.toString("base64"),
            index: sig.index,
        })),
        consistencyLevel: vaa.consistencyLevel,
        guardianSetIndex: vaa.guardianSetIndex,
    };
}
function deserializeVaa(vaa) {
    return {
        sequence: BigInt(vaa.sequence),
        hash: Buffer.from(vaa.hash, "base64"),
        emitterChain: vaa.emitterChain,
        emitterAddress: Buffer.from(vaa.emitterAddress, "hex"),
        payload: Buffer.from(vaa.payload, "base64"),
        nonce: vaa.nonce,
        timestamp: vaa.timestamp,
        version: vaa.version,
        guardianSignatures: vaa.guardianSignatures.map((sig) => ({
            signature: Buffer.from(sig.signature, "base64"),
            index: sig.index,
        })),
        consistencyLevel: vaa.consistencyLevel,
        guardianSetIndex: vaa.guardianSetIndex,
    };
}
const defaultOptions = {
    attempts: 3,
    redis: {},
    queueName: "relays",
    concurrency: 3,
    maxCompletedQueueSize: 10000,
    maxFailedQueueSize: 10000,
};
class RedisStorage {
    logger;
    vaaQueue;
    registry;
    workerId;
    worker;
    prefix;
    redis;
    metrics;
    opts;
    constructor(opts) {
        this.opts = Object.assign({}, defaultOptions, opts);
        // ensure redis is defined
        if (!this.opts.redis) {
            this.opts.redis = {};
        }
        this.opts.redis.maxRetriesPerRequest = null; //Added because of: DEPRECATION WARNING! Your redis options maxRetriesPerRequest must be null. On the next versions having this settings will throw an exception
        this.prefix = `{${this.opts.namespace ?? this.opts.queueName}}`;
        this.redis =
            this.opts.redisClusterEndpoints !== undefined &&
                this.opts.redisClusterEndpoints.length > 0
                ? new ioredis_1.Redis.Cluster(this.opts.redisClusterEndpoints, this.opts.redisCluster)
                : new ioredis_1.Redis(this.opts.redis);
        // TODO: consider using a queue per chain
        this.vaaQueue = new bullmq_1.Queue(this.opts.queueName, {
            defaultJobOptions: {
                removeOnComplete: this.opts.maxCompletedQueueSize,
                removeOnFail: this.opts.maxFailedQueueSize,
            },
            prefix: this.prefix,
            connection: this.redis,
        });
        const { metrics, registry } = (0, storage_metrics_js_1.createStorageMetrics)();
        this.metrics = metrics;
        this.registry = registry;
    }
    getPrefix() {
        return [this.prefix, this.opts.queueName].join(":");
    }
    async addVaaToQueue(vaaBytes) {
        const startTime = Date.now();
        const parsedVaa = (0, wormhole_sdk_1.parseVaa)(vaaBytes);
        const id = this.vaaId(parsedVaa);
        const idWithoutHash = id.substring(0, id.length - 6);
        this.logger?.debug(`Adding VAA to queue`, {
            emitterChain: parsedVaa.emitterChain,
            emitterAddress: parsedVaa.emitterAddress.toString("hex"),
            sequence: parsedVaa.sequence.toString(),
        });
        const retryStrategy = this.opts.exponentialBackoff
            ? {
                backoff: {
                    type: "custom",
                },
            }
            : undefined;
        const job = await this.vaaQueue.add(idWithoutHash, {
            parsedVaa: serializeVaa(parsedVaa),
            vaaBytes: vaaBytes.toString("base64"),
        }, {
            jobId: id,
            removeOnFail: 50000,
            attempts: this.opts.attempts,
            ...retryStrategy,
        });
        return {
            attempts: 0,
            data: { vaaBytes, parsedVaa },
            id: job.id,
            name: job.name,
            log: job.log.bind(job),
            receivedAt: startTime,
            updateProgress: job.updateProgress.bind(job),
            maxAttempts: this.opts.attempts,
        };
    }
    startWorker(handleJob) {
        this.logger?.debug(`Starting worker for queue: ${this.opts.queueName}. Prefix: ${this.prefix}.`);
        // Use user provided backoff function if available, otherwise use xlabs default
        let backOffFunction = undefined;
        if (this.opts.exponentialBackoff?.backOffFn) {
            backOffFunction = this.opts.exponentialBackoff.backOffFn;
        }
        else {
            backOffFunction = (attemptsMade) => {
                const exponentialDelay = Math.pow(2, attemptsMade) *
                    (this.opts.exponentialBackoff?.baseDelayMs || 1000);
                return Math.min(exponentialDelay, this.opts.exponentialBackoff?.maxDelayMs || 3600000);
            };
        }
        const workerSettings = this.opts.exponentialBackoff
            ? {
                settings: {
                    backoffStrategy: backOffFunction,
                },
            }
            : undefined;
        this.worker = new bullmq_1.Worker(this.opts.queueName, async (job) => {
            let parsedVaa = job.data?.parsedVaa;
            if (parsedVaa) {
                this.logger?.debug(`Starting job: ${job.id}`, {
                    emitterChain: parsedVaa.emitterChain,
                    emitterAddress: parsedVaa.emitterAddress.toString("hex"),
                    sequence: parsedVaa.sequence.toString(),
                });
            }
            else {
                this.logger?.debug("Received job with no parsedVaa");
            }
            const vaaBytes = Buffer.from(job.data.vaaBytes, "base64");
            const relayJob = {
                attempts: job.attemptsMade,
                data: {
                    vaaBytes,
                    parsedVaa: (0, wormhole_sdk_1.parseVaa)(vaaBytes),
                },
                id: job.id,
                maxAttempts: this.opts.attempts,
                name: job.name,
                receivedAt: job.timestamp,
                log: job.log.bind(job),
                updateProgress: job.updateProgress.bind(job),
            };
            await job.log(`processing by..${this.workerId}`);
            await handleJob(relayJob);
            return;
        }, {
            prefix: this.prefix,
            connection: this.redis,
            concurrency: this.opts.concurrency,
            ...workerSettings,
        });
        this.workerId = this.worker.id;
        this.spawnGaugeUpdateWorker();
    }
    async stopWorker() {
        await this.worker?.close();
        this.worker = undefined;
    }
    async spawnGaugeUpdateWorker(ms = 5000) {
        while (this.worker !== undefined) {
            await this.updateGauges();
            await (0, utils_js_1.sleep)(ms);
        }
    }
    storageKoaUI(path) {
        // UI
        const serverAdapter = new koa_1.KoaAdapter();
        serverAdapter.setBasePath(path);
        (0, api_1.createBullBoard)({
            queues: [new bullMQAdapter_js_1.BullMQAdapter(this.vaaQueue)],
            serverAdapter: serverAdapter,
        });
        return serverAdapter.registerPlugin();
    }
    vaaId(vaa) {
        const emitterAddress = vaa.emitterAddress.toString("hex");
        const hash = vaa.hash.toString("base64").substring(0, 5);
        let sequence = vaa.sequence.toString();
        return `${vaa.emitterChain}/${emitterAddress}/${sequence}/${hash}`;
    }
    async updateGauges() {
        const { active, delayed, waiting, failed } = await this.vaaQueue.getJobCounts();
        this.metrics.activeGauge.labels({ queue: this.vaaQueue.name }).set(active);
        this.metrics.delayedGauge
            .labels({ queue: this.vaaQueue.name })
            .set(delayed);
        this.metrics.waitingGauge
            .labels({ queue: this.vaaQueue.name })
            .set(waiting);
        this.metrics.failedGauge.labels({ queue: this.vaaQueue.name }).set(failed);
    }
}
exports.RedisStorage = RedisStorage;
//# sourceMappingURL=redis-storage.js.map