import { RelayerApp } from "./application.js";
import { logging, providers, sourceTx, spawnMissedVaaWorker, stagingArea, tokenBridgeContracts, wallets, } from "./middleware/index.js";
import { RedisStorage, } from "./storage/redis-storage.js";
import { mergeDeep } from "./utils.js";
import { defaultLogger } from "./logging.js";
import { BullMQAdapter } from "@bull-board/api/bullMQAdapter.js";
import { KoaAdapter } from "@bull-board/koa";
import { createBullBoard } from "@bull-board/api";
import { Registry } from "prom-client";
const defaultStdOpts = {
    spyEndpoint: "localhost:7073",
    workflows: {
        retries: 3,
    },
    fetchSourceTxhash: true,
    logger: defaultLogger,
};
export class StandardRelayerApp extends RelayerApp {
    store;
    mergedRegistry;
    constructor(env, opts) {
        // take logger out before merging because of recursive call stack
        const logger = opts.logger ?? defaultLogger;
        delete opts.logger;
        // now we can merge
        const options = mergeDeep({}, [
            defaultStdOpts,
            opts,
        ]);
        const { privateKeys, tokensByChain, name, spyEndpoint, redis, redisCluster, redisClusterEndpoints, wormholeRpcs, retryBackoffOptions, maxCompletedQueueSize, maxFailedQueueSize, } = options;
        super(env, options);
        this.store = new RedisStorage({
            redis,
            redisClusterEndpoints,
            redisCluster,
            attempts: options.workflows?.retries ?? 3,
            namespace: name,
            queueName: `${name}-relays`,
            exponentialBackoff: retryBackoffOptions,
            maxCompletedQueueSize,
            maxFailedQueueSize,
        });
        // this is always true for standard relayer app, but I'm adding this if
        // to make the dependency explicit
        if (this.store) {
            spawnMissedVaaWorker(this, {
                namespace: name,
                registry: this.store.registry,
                logger,
                redis,
                redisCluster,
                redisClusterEndpoints,
                wormholeRpcs,
                concurrency: options.missedVaaOptions?.concurrency,
                checkInterval: options.missedVaaOptions?.checkInterval,
                fetchVaaRetries: options.missedVaaOptions?.fetchVaaRetries,
                vaasFetchConcurrency: options.missedVaaOptions?.vaasFetchConcurrency,
                storagePrefix: this.store.getPrefix(),
                startingSequenceConfig: options.missedVaaOptions?.startingSequenceConfig,
                forceSeenKeysReindex: options.missedVaaOptions?.forceSeenKeysReindex,
            });
        }
        this.mergedRegistry = Registry.merge([
            this.store.registry,
            super.metricsRegistry,
        ]);
        this.spy(spyEndpoint);
        this.useStorage(this.store);
        this.logger(logger);
        this.use(logging(logger));
        this.use(providers(options.providers, Object.keys(privateKeys ?? {})));
        // You need valid private keys to turn on the wallet middleware
        if (privateKeys !== undefined && Object.keys(privateKeys).length > 0) {
            this.use(wallets(env, {
                logger,
                namespace: name,
                privateKeys,
                tokensByChain,
                metrics: { enabled: true, registry: this.metricsRegistry },
            }));
        }
        this.use(tokenBridgeContracts());
        this.use(stagingArea({
            namespace: name,
            redisCluster,
            redis,
            redisClusterEndpoints,
        }));
        if (options.fetchSourceTxhash) {
            this.use(sourceTx());
        }
    }
    /**
     * Registry with prometheus metrics exported by the relayer.
     * Metrics include:
     * - active_workflows: Number of workflows currently running
     * - delayed_workflows: Number of worklows which are scheduled in the future either because they were scheduled that way or because they failed.
     * - waiting_workflows: Workflows waiting for a worker to pick them up.
     * - worklow_processing_duration: Processing time for completed jobs (processing until completed)
     * - workflow_total_duration: Processing time for completed jobs (processing until completed)
     */
    get metricsRegistry() {
        return this.mergedRegistry;
    }
    /**
     * A UI that you can mount in a KOA app to show the status of the queue / jobs.
     * @param path
     */
    storageKoaUI(path) {
        // UI
        const serverAdapter = new KoaAdapter();
        serverAdapter.setBasePath(path);
        createBullBoard({
            queues: [new BullMQAdapter(this.store.vaaQueue)],
            serverAdapter: serverAdapter,
        });
        return serverAdapter.registerPlugin();
    }
}
//# sourceMappingURL=application-standard.js.map