import { Redis } from "ioredis";
import { createPool } from "generic-pool";
import { defaultWormholeRpcs, defaultWormscanUrl, } from "../../application.js";
import { mapConcurrent, sleep } from "../../utils.js";
import { initMetrics } from "./metrics.js";
import { calculateSequenceStats, updateMetrics, } from "./helpers.js";
import { checkForMissedVaas, refreshSeenSequences, registerEventListeners, } from "./check.js";
import { tryGetExistingFailedSequences, tryGetLastSafeSequence, trySetLastSafeSequence, } from "./storage.js";
import { WormholescanClient, } from "../../rpc/wormholescan-client.js";
const DEFAULT_PREFIX = "MissedVaaWorkerV3";
let metrics;
export async function spawnMissedVaaWorker(app, opts) {
    opts.wormholeRpcs = opts.wormholeRpcs ?? defaultWormholeRpcs[app.env];
    opts.wormscanUrl = opts.wormscanUrl ?? defaultWormscanUrl[app.env];
    if (!metrics) {
        metrics = opts.registry ? initMetrics(opts.registry) : {};
    }
    if (!opts.maxLookAhead && opts.maxLookAhead !== 0) {
        opts.maxLookAhead = 10;
    }
    const redisPool = createRedisPool(opts);
    const wormholescan = new WormholescanClient(new URL(opts.wormscanUrl), {
        maxDelay: 60000,
        noCache: true,
    });
    if (!app.filters.length) {
        opts.logger?.warn("Missed VAA Worker: No filters found, retrying in 100ms...");
        setTimeout(() => {
            spawnMissedVaaWorker(app, opts);
        }, 100);
        return;
    }
    const filters = app.filters.map(filter => {
        return {
            emitterChain: filter.emitterFilter.chainId,
            emitterAddress: filter.emitterFilter.emitterAddress,
        };
    });
    const prefix = opts.storagePrefix || DEFAULT_PREFIX;
    if (opts.storagePrefix) {
        await refreshSeenSequences(redisPool, filters, opts);
    }
    registerEventListeners(app, redisPool, opts, prefix);
    while (true) {
        opts.logger?.info(`Missed VAA middleware run starting...`);
        await mapConcurrent(filters, filter => redisPool.use(async (redis) => {
            const startTime = Date.now();
            const filterLogger = opts.logger?.child({
                emitterChain: filter.emitterChain,
                emitterAddress: filter.emitterAddress,
            });
            try {
                const results = await runMissedVaaCheck(filter, redis, app.processVaa.bind(app), opts, prefix, wormholescan, filterLogger);
                updateMetrics(metrics, filter, startTime, false, results.missedVaas, results.sequenceStats);
            }
            catch (error) {
                filterLogger?.error(`Error checking for missed vaas for filter: ${JSON.stringify(filter)}}`, error);
                updateMetrics(metrics, filter, startTime, true);
            }
        }));
        await sleep(opts.checkInterval || 30000);
    }
}
export async function runMissedVaaCheck(filter, redis, processVaa, opts, storagePrefix, wormholescan, logger) {
    const { emitterChain, emitterAddress } = filter;
    const previousSafeSequence = await tryGetLastSafeSequence(redis, storagePrefix, emitterChain, emitterAddress, logger);
    const missedVaas = await checkForMissedVaas(filter, redis, processVaa, opts, storagePrefix, wormholescan, previousSafeSequence, logger);
    // TODO: this is an ugly way to handle the error
    const failedToFetchSequencesOrError = await tryGetExistingFailedSequences(redis, filter, storagePrefix);
    if (!Array.isArray(failedToFetchSequencesOrError)) {
        logger?.error(`Failed to get existing failed sequences from redis. Error: `, failedToFetchSequencesOrError);
    }
    else if (failedToFetchSequencesOrError.length) {
        logger?.warn(`Found sequences that we failed to get from wormhole-rpc. Sequences: ` +
            JSON.stringify(failedToFetchSequencesOrError));
    }
    else {
        logger?.debug("No previous failed sequences found.");
    }
    const failedToFetchSequences = Array.isArray(failedToFetchSequencesOrError)
        ? failedToFetchSequencesOrError
        : null;
    const sequenceStats = calculateSequenceStats(missedVaas, failedToFetchSequences, previousSafeSequence?.toString());
    const { lastSafeSequence, lastSeenSequence, firstSeenSequence } = sequenceStats;
    if (!previousSafeSequence ||
        previousSafeSequence?.toString() !== String(lastSafeSequence)) {
        logger?.debug(`No missing sequences found up to sequence ${lastSafeSequence}. Setting as last sequence`);
        await trySetLastSafeSequence(redis, storagePrefix, emitterChain, emitterAddress, lastSafeSequence);
    }
    const vaasFound = missedVaas.missingSequences.length + missedVaas.lookAheadSequences.length;
    logger?.info(`Finished missed VAAs check. Found: ${vaasFound}. Enable debug to see a log with precise results.`);
    logger?.debug(`Finished missed vaas check. Results: ${JSON.stringify({
        missedVaas,
        lastSafeSequence,
        lastSeenSequence,
        firstSeenSequence,
    })}}`);
    return { missedVaas, sequenceStats };
}
function createRedisPool(opts) {
    const factory = {
        create: async function () {
            const redis = opts.redisCluster
                ? new Redis.Cluster(opts.redisClusterEndpoints, opts.redisCluster)
                : new Redis(opts.redis);
            // TODO: metrics.missed_vaa_redis_open_connections.inc();
            return redis;
        },
        destroy: async function (redis) {
            // TODO: metrics.missed_vaa_redis_open_connections.dec();
        },
    };
    const poolOpts = {
        min: 5,
        max: 15,
        autostart: true,
    };
    return createPool(factory, poolOpts);
}
//# sourceMappingURL=worker.js.map