import { CHAIN_ID_SEI, CHAIN_ID_SOLANA, CHAIN_ID_SUI, CHAIN_ID_TO_NAME, } from "@certusone/wormhole-sdk";
import { spawnWalletWorker } from "./wallet.worker.js";
import { Queue } from "@datastructures-js/queue";
import { startWalletManagement } from "./wallet-management.js";
function makeExecuteFunc(actionQueues, pluginName, logger) {
    // push action onto actionQueue and have worker reject or resolve promise
    const func = (chainId, f) => {
        return new Promise((resolve, reject) => {
            const maybeQueue = actionQueues.get(chainId);
            if (!maybeQueue) {
                logger?.error(`Error making execute function. Unsupported chain: ${chainId}`);
                return reject("Chain not supported");
            }
            maybeQueue.enqueue({
                action: { chainId, f },
                pluginName,
                resolve,
                reject,
            });
        });
    };
    func.onSolana = (f) => func(CHAIN_ID_SOLANA, f);
    func.onSui = (f) => func(CHAIN_ID_SUI, f);
    func.onSei = (f) => func(CHAIN_ID_SEI, f);
    func.onEVM = (chainId, f) => func(chainId, f);
    return func;
}
export function wallets(env, opts) {
    const workerInfoMap = new Map(Object.entries(opts.privateKeys).map(([chainIdStr, keys]) => {
        //TODO update for all ecosystems
        let chainId = Number(chainIdStr);
        const workerInfos = keys.map((key, id) => ({
            id,
            targetChainId: chainId,
            targetChainName: CHAIN_ID_TO_NAME[chainId],
            walletPrivateKey: key,
        }));
        return [chainId, workerInfos];
    }));
    if (opts.metrics) {
        startWalletManagement(env, opts.privateKeys, opts.tokensByChain, opts.metrics, opts.logger);
    }
    let executeFunction;
    return async (ctx, next) => {
        if (!executeFunction) {
            ctx.logger?.debug(`Initializing wallets...`);
            const actionQueues = new Map();
            for (const [chain, workerInfos] of workerInfoMap.entries()) {
                const actionQueue = new Queue();
                actionQueues.set(chain, actionQueue);
                workerInfos.forEach(info => spawnWalletWorker(actionQueue, ctx.providers, info, opts.logger));
            }
            executeFunction = makeExecuteFunc(actionQueues, opts.namespace ?? "default", opts.logger);
            ctx.logger?.debug(`Initialized wallets`);
        }
        ctx.logger?.debug("wallets attached to context");
        ctx.wallets = executeFunction;
        await next();
    };
}
//# sourceMappingURL=wallet.middleware.js.map