import * as wormholeSdk from "@certusone/wormhole-sdk";
import { isChain, isEVMChain, parseVaa, } from "@certusone/wormhole-sdk";
import { bech32 } from "bech32";
import { deriveWormholeEmitterKey } from "@certusone/wormhole-sdk/lib/cjs/solana/wormhole/index.js";
import { zeroPad } from "ethers/lib/utils.js";
import { ethers } from "ethers";
import { inspect } from "util";
export function encodeEmitterAddress(chainId, emitterAddressStr) {
    if (chainId === wormholeSdk.CHAIN_ID_SOLANA ||
        chainId === wormholeSdk.CHAIN_ID_PYTHNET) {
        return deriveWormholeEmitterKey(emitterAddressStr)
            .toBuffer()
            .toString("hex");
    }
    if (wormholeSdk.isCosmWasmChain(chainId)) {
        return Buffer.from(zeroPad(bech32.fromWords(bech32.decode(emitterAddressStr).words), 32)).toString("hex");
    }
    if (wormholeSdk.isEVMChain(chainId)) {
        return wormholeSdk.getEmitterAddressEth(emitterAddressStr);
    }
    if (wormholeSdk.CHAIN_ID_ALGORAND === chainId) {
        return wormholeSdk.getEmitterAddressAlgorand(BigInt(emitterAddressStr));
    }
    if (wormholeSdk.CHAIN_ID_NEAR === chainId) {
        return wormholeSdk.getEmitterAddressNear(emitterAddressStr);
    }
    if (wormholeSdk.CHAIN_ID_SUI === chainId) {
        return strip0x(emitterAddressStr);
    }
    throw new Error(`Unrecognized wormhole chainId ${chainId}`);
}
export const strip0x = (str) => str.startsWith("0x") ? str.substring(2) : str;
export function sleep(ms) {
    return new Promise((resolve, reject) => setTimeout(resolve, ms));
}
/**
 * Simple object check.
 * @param item
 * @returns {boolean}
 */
export function isObject(item) {
    return item && typeof item === "object" && !Array.isArray(item);
}
export function parseVaaWithBytes(bytes) {
    const parsedVaa = parseVaa(bytes);
    const id = {
        emitterChain: parsedVaa.emitterChain,
        emitterAddress: parsedVaa.emitterAddress.toString("hex"),
        sequence: parsedVaa.sequence.toString(),
    };
    return { ...parsedVaa, bytes, id };
}
/**
 * Deep merge two objects.
 * @param target
 * @param ...sources
 */
export function mergeDeep(target, sources, maxDepth = 10) {
    if (!sources.length || maxDepth === 0) {
        // @ts-ignore
        return target;
    }
    const source = sources.shift();
    if (isObject(target) && isObject(source)) {
        for (const key in source) {
            const sourceProp = source[key];
            if (isObject(sourceProp)) {
                const targetProp = target[key] || {};
                // We need to cast because the narrowed type for the key is lost in the object.
                if (target[key] === undefined)
                    target[key] = targetProp;
                mergeDeep(targetProp, [sourceProp], maxDepth - 1);
            }
            else {
                Object.assign(target, { [key]: source[key] });
            }
        }
    }
    return mergeDeep(target, sources, maxDepth);
}
export const second = 1000;
export const minute = 60 * second;
export const hour = 60 * minute;
export class EngineError extends Error {
    args;
    constructor(msg, args) {
        super(msg);
        this.args = args;
    }
}
export function maybeConcat(...arrs) {
    return arrs.flatMap(arr => (arr ? arr : []));
}
export function nnull(x, errMsg) {
    if (x === undefined || x === null) {
        throw new Error("Found unexpected undefined or null. " + errMsg);
    }
    return x;
}
export function assertStr(x, fieldName) {
    if (typeof x !== "string") {
        throw new EngineError(`Expected field to be integer, found ${x}`, {
            fieldName,
        });
    }
    return x;
}
export function assertInt(x, fieldName) {
    if (!Number.isInteger(Number(x))) {
        throw new EngineError(`Expected field to be integer, found ${x}`, {
            fieldName,
        });
    }
    return x;
}
export function assertArray(x, name, elemsPred) {
    if (!Array.isArray(x) || (elemsPred && !x.every(elemsPred))) {
        throw new EngineError(`Expected value to be array, found ${x}`, {
            name,
        });
    }
    return x;
}
export function assertBool(x, fieldName) {
    if (x !== false && x !== true) {
        throw new EngineError(`Expected field to be boolean, found ${x}`, {
            fieldName,
        });
    }
    return x;
}
export function wormholeBytesToHex(address) {
    return ethers.utils.hexlify(address).replace("0x", "");
}
export function assertEvmChainId(chainId) {
    if (!isEVMChain(chainId)) {
        throw new EngineError("Expected number to be valid EVM chainId", {
            chainId,
        });
    }
    return chainId;
}
export function assertChainId(chainId) {
    if (!isChain(chainId)) {
        throw new EngineError("Expected number to be valid chainId", { chainId });
    }
    return chainId;
}
export function dbg(x, msg) {
    if (msg) {
        console.log(msg);
    }
    console.log(x);
    return x;
}
export async function mapConcurrent(arr, fn, concurrency = 5) {
    const pendingArgs = [...arr];
    async function evaluateNext() {
        if (pendingArgs.length === 0)
            return;
        const args = pendingArgs.shift();
        await fn(args);
        // If any pending promise is resolved, then evaluate next
        await evaluateNext();
    }
    // Promises that will be executed parallely, with a maximum of `concurrency` at a time
    const promises = new Array(concurrency).fill(0).map(evaluateNext);
    await Promise.all(promises);
}
export function printError(error) {
    if (error instanceof Error) {
        return `${error?.stack || error.message}`;
    }
    // Prints nested properties until a depth of 2 by default.
    return inspect(error);
}
export function min(lhs, rhs) {
    return lhs < rhs ? lhs : rhs;
}
export function max(lhs, rhs) {
    return lhs < rhs ? rhs : lhs;
}
//# sourceMappingURL=utils.js.map